using System.Reflection.Emit;
using Micro.Interfaces.SymTab;

namespace Micro.Implementation
{
    /// Symbol Table Nodes:
    /// Every named object in a program is stored in an Symbol node.
    /// Every scope has a list of Symbols declared within it.
    internal class Symbol : ISymbol
    {
        private SymbolKind kind;
        private string name;
        private IStruct type;
        private ISymbol next;
        private int val;
        private int adr;
        private int nArgs;
        private int nLocs;
        private ISymbol locals;
        private MethodBuilder meth;

        public Symbol(SymbolKind kind, string name, IStruct type)
        {
            this.kind = kind;
            this.name = name;
            this.type = type;
        }

        public SymbolKind Kind
        {
            get { return kind; }
            set { kind = value; }
        }

        public string Name
        {
            get { return name; }
            set { name = value; }
        }

        public IStruct Type
        {
            get { return type; }
            set { type = value; }
        }

        /// next Symbol node in this scope
        public ISymbol Next
        {
            get { return next; }
            set { next = value; }
        }

        /// Const: value;
        public int Val
        {
            get { return val; }
            set { val = value; }
        }

        /// Field, Arg, Local: order of declaration in scope
        public int Adr
        {
            get { return adr; }
            set { adr = value; }
        }

        /// Func: # of arguments
        public int NArgs
        {
            get { return nArgs; }
            set { nArgs = value; }
        }

        /// Func: # of local variables
        public int NLocs
        {
            get { return nLocs; }
            set { nLocs = value; }
        }

        /// Func: arguments, then local variables;
        public ISymbol Locals
        {
            get { return locals; }
            set { locals = value; }
        }

        /// this field is necessary for the building the assembly
        /// with System.Reflection.Emit
        /// Method: builder for metadata and CIL
        public MethodBuilder Meth
        {
            get { return meth; }
            set { meth = value; }
        }

        public override bool Equals(object o)
        {
            if (this == o)
            {
                // same object
                return true;
            }

            Symbol sym = o as Symbol;

            if (sym == null || kind != sym.kind || name != sym.name ||
                !type.Equals(sym.type))
            {
                return false;
            }

            switch (kind)
            {
                case SymbolKind.Const:
                    return val == sym.val;
                case SymbolKind.Arg:
                case SymbolKind.Local:
                    return adr == sym.adr;
                case SymbolKind.Func:
                    return nArgs == sym.nArgs && nLocs == sym.nLocs &&
                           EqualsCompleteList(locals, sym.locals);
            }

            return true;
        }

        public override int GetHashCode()
        {
            return (int)kind;
        }

        /// Compare complete Symbol node lists.
        private static bool EqualsCompleteList(ISymbol sym1, ISymbol sym2)
        {
            if (sym1 == sym2)
            {
                // same object
                return true;
            }

            while (sym1 != null && sym1.Equals(sym2))
            {
                sym1 = sym1.Next;
                sym2 = sym2.Next;
            }

            return sym1 == null && sym2 == null;
        }
    }
}